//   Copyright (C) 2005 The Regents of the University of California.
//   All Rights Reserved.
//
//   Permission to use, copy, modify, and distribute this software and its
//   documentation for educational, research and non-profit purposes,
//   without fee, and without a written agreement is hereby granted,
//   provided that the above copyright notice, this paragraph and the
//   following three paragraphs appear in all copies.
//
//   Permission to incorporate this software into commercial products may
//   be obtained by contacting the University of California. For
//   information about obtaining such a license contact:
//
//   Chrisanna Waldrop
//   Copyright Officer
//   805-893-7773
//   waldrop@research.ucsb.edu
//
//   This software program and documentation are copyrighted by The Regents
//   of the University of California. The software program and
//   documentation are supplied "as is", without any accompanying services
//   from The Regents. The Regents does not warrant that the operation of
//   the program will be uninterrupted or error-free. The end-user
//   understands that the program was developed for research purposes and
//   is advised not to rely exclusively on the program for any reason.
//
//   IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
//   FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
//   INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
//   ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
//   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
//   CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//   A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
//   BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE
//   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
package edu.ucsb.nmsl.autocap;

import java.util.Iterator;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Arrays;
import java.util.List;
import java.util.ListIterator;
import java.util.Stack;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import edu.ucsb.nmsl.tools.*;
import edu.cmu.sphinx.result.Result;

/**
 * This class is reponsible for aligning captions from utterances recognized by 
 * Sphinx. This is done in essentially two steps, first find longest common 
 * substrings and bursts, then estimating captions that are not recognized. 
 * These two steps will eventually be put in their own classes. This is a proof
 * of concept implementation, not a final implementation.
 *
 * @author  Allan Knight
 * @version 1.0
 *
 */
public class CaptionAligner
{
  /** Collection of transcripts of type Caption. */
  Transcript transcript = null;

  /** Linked list of all words in text, lowercase and punctuation removed. */
  LinkedList text = null;

  /** Linked list of bursts from the utterances in the transcript. */
  List bursts = null;

  /** RegEx pattern for matching the start of a time from Sphinx. */
  Pattern startPat = null;

  /** RegEx pattern for matching the end of a time from Sphinx. */
  Pattern endPat = null;

  /** Holds the amount of time speaking during bursts. */
  double speakingTime = 0.0;

  /** Holds number of words from transcript matched. */
  double wordsMatched = 0.0;

  /** Holds total number of words in transcript. */
  double totalWords = 0.0;

  /** Holds start time of aligning so time can be measured. */
  double startTime = 0.0;

  /** Holds finish time of aligning so time can be measured. */
  double finishTime = 0.0;

  /** Holds total number of sentences in transcript. */
  double totalSentences = 0.0;

  /** Holds number of sentences with some coverage. */
  double sentencesCovered = 0.0;

  /** Collects all the speaking rates for analysis */
  DataSetStatistic SpeakingRate = new DataSetStatistic("Speaking Rate Data");
  
  /** Collects all speaking rates for unrecognized word bursts. */
  DataSetStatistic UncoveredRate = 
    new DataSetStatistic("Uncovered Speaking Rate Data");
    
  /** */
  DataSetStatistic UncoveredWords = 
    new DataSetStatistic("Uncovered Words Data");

  /** Burst length for alignment. Any burst must have at least this many words
      in it. */
  static final int BURST_MIN_LENGTH = 4;
  
  /** Linked list of all recognized words and their time-stamps as returned by
      Sphinx. */
  LinkedList Timed = new LinkedList();
  
  /** Linked list of all words in the trancript. This data memeber along with
      RawText is used to calculate the LCS during the Alignment phase of 
      AutoCap. */
  LinkedList Raw   = new LinkedList();
  
  /** Raw text of all recognized words */
  String     RawText = new String();

  /**
   * This constructor takes in a DOM Document that contains captioning
   * information for a particular presentation. Extracts sentences from
   * document and then compile patterns used in other methods.
   *
   * @param t The transcript for which alignment of captions will be performed.
   *
   */
 public CaptionAligner(Transcript t)
  {
    // Allocate memory for burst linkedlist
    bursts = new LinkedList();

    // Extract all the sentences from the transcript
    extractSentences(t);
    
    // Compile time patterns so we can extract times for utterances.
    startPat = Pattern.compile("\\((\\d+\\.\\d+)");
    endPat = Pattern.compile("\\,(\\d+\\.\\d+)\\)");
  }

  /**
   * Extracts sentences from XML document that contains captioning information
   * for a presentation. All sentences are in the text attribute of the
   * language field. All other fields are ignored.
   *
   * @param t - Transcript object that contains captions.
   *
   */
  protected void extractSentences(Transcript t)
  {
    // Allocate the LinkedList to hold each caption of the transcript
    transcript = t;

    // Allocate LinkedList to hold each word of the transcript
    text = new LinkedList(Arrays.asList(transcript.getAsText().split("\\s+")));
    totalWords = text.size();
    RawText = transcript.getAsText();
  }

  /**
   * Adds utterances as they come in from Sphinx. Finds the bursts and adds
   * them to the burst LinkedList. These are processed later for estimating
   * start times of each caption.
   *
   * @param r - The Result object containing the most recent utterance.
   *
   * @return true if successful, false otherwise.
   *
   */
  public boolean addUtterance(Result r)
  {
    if(r.getBestResultNoFiller().length() <= 0)
      return false;

    double lastStart = 0.0;
    double lastFinish = 0.0;

    // See if we've started
    if (startTime == 0.0)
    {
      startTime = (double) System.currentTimeMillis();
    }

    // Get the best result from Sphinx
    LinkedList raw = new LinkedList
      (
        Arrays.asList(r.getBestResultNoFiller().split("\\s+"))
      );

    // Kludge. Sometimes getTimedBestResult throws a NullPointerException.
    // So I just catch it and return false if it happens. The utterance is
    // then skipped because we can't get any timing information from it.
    LinkedList timed;
    try
    {
      timed = new LinkedList
    	(
    	  Arrays.asList(r.getTimedBestResult(false, true).split("\\s+"))
    	);
    } 
    catch (NullPointerException e)
    {
      System.err.println("getTimedBestResult threw a NullPointerException");
      return false;
    }

    // Make sure both the raw and timed have the same number of elements
    if(raw.size() != timed.size())
    {
      for
      (
        ListIterator i = raw.listIterator(), j = timed.listIterator(); 
	      i.hasNext();
      )
      {
      	String rawString = (String)i.next();
      	if(!j.hasNext())
      	{
      	  j.add(rawString + "(" + lastStart + "," + lastFinish + ")");
      	}
      	else
      	{
      	  String timeString = (String)j.next();
      	  if(!timeString.startsWith(rawString))
      	  {
      	    j.add(rawString + "(" + lastStart + "," + lastFinish + ")");
      	    if(j.hasNext())
      	    j.next();
      	  }
      	  Matcher m = startPat.matcher(timeString);
      	  m.find();
      	  lastStart = Double.parseDouble(m.group(1));
            
      	  m = endPat.matcher(timeString);
      	  m.find();
      	  lastFinish = Double.parseDouble(m.group(1));
        }
      }
    }

    Raw.addAll(raw);
    Timed.addAll(timed);

    // We can take our finish time here each time, then when we're done we
    // can calculate the total processing time.
    finishTime = (double) System.currentTimeMillis();

    return true;
 }

  /**
   * 
   * This method collects all the burst of minimum lenght of recognized, and
   * therefore time-stamped words. These bursts are used to actually extract the
   * time when words were spoken in the video.
   *
   */
  protected boolean collectBursts()
  {
    LinkedList burst = new LinkedList();

    // Get the Longest common subsequence from utterance and transcript
    LinkedList spoken = new LinkedList(LCS.getLCS(Raw, text));

    // If the LCSS is smaller than the burst size, there's no need to continue
    if (spoken.size() < BURST_MIN_LENGTH)
      return true;

    // While there are more words in the spoken list
    while(!spoken.isEmpty())
    {
      // Find the first occurence of spoken[i] in text
      int j = text.indexOf(spoken.getFirst());

      // Remove everything upto jth element
      for(int idx = 0; idx <= j; ++idx)
      {
	      text.removeFirst();
      }
      burst.add(spoken.getFirst());
      spoken.removeFirst();

      while(!spoken.isEmpty() && text.getFirst().equals(spoken.getFirst()))
      {
      	burst.add(spoken.getFirst());
      	spoken.removeFirst();
      	text.removeFirst();
      }

      if(burst.size() >= BURST_MIN_LENGTH)
      {
      	Caption c = createTimedCaption(burst, Raw, Timed);
      	SpeakingRate.addDataPoint
      	(
      	  (c.getFinishTime() - c.getTime()) / burst.size()
      	);
      	wordsMatched += burst.size();
      	bursts.add(c);
      }
      burst.clear();
    }

    // If you got hear all is well
    return true;
  }

  /**
   * Aligns the captions that we have collected. That means, it iterates
   * through each word and marks those it can identify with the correct time.
   * This method is called once all all the audio has been processed. This
   * method should eventually be replaced by a class that is plugged in to
   * provide this functionality.
   *
   * @return A Transcript with all captions aligned to approximately where they
   *         were spoken in a given video or audio file.
   *
   */
  public Transcript getAlignedCaptions()
  {
    collectBursts();

    // Create iterators to iterate through each word of each caption
    ListIterator burstItr = bursts.listIterator(0);

    if(!burstItr.hasNext())
    {
      return null;
    }

    Caption burstCaption = (Caption) burstItr.next();
    ListIterator bWordItr = burstCaption.listIterator(0);
    Caption.CaptionChunk bChunk = (Caption.CaptionChunk) bWordItr.next();
    
    ListIterator capStart = null;
    ListIterator wordStart = null;

    // Create stacks so we can push the words on before we reach a minimum
    // burst size.
    Stack bStack = new Stack();
    Stack tStack = new Stack();


    int word_count = 0;


    // Iterate through each word of a transcript and look for bursts. First
    // we need an iterator for each caption.
    for (Iterator transItr = transcript.captionIterator(0); transItr.hasNext();)
    {
      // Now we need an iterator for each word in the iterator.
      Caption transCaption = (Caption) transItr.next();
      for
      (
        ListIterator tWordItr = transCaption.listIterator(0);
	      tWordItr.hasNext();
      )
      {
        // Get the next caption chunk. Most likely a word.
        Caption.CaptionChunk tChunk = (Caption.CaptionChunk) tWordItr.next();

        // Look for the words in the burst to be equal to those in the
        // transcript.
        if (tChunk.getCaption().equals(bChunk.getCaption()))
        {
          // If this is the first words to match in burst, save the starting
          // position of the burst. That is the caption and word.
          if (word_count == 0)
          {
            capStart = bursts.listIterator(burstItr.nextIndex());
            wordStart = burstCaption.listIterator(bWordItr.nextIndex() - 1);
          }

          // Increment the word count since we matched words
          ++word_count;

          // If we're already in a burst, then set the start time and finish
          // time for the word
          if (word_count > BURST_MIN_LENGTH)
          {
            tChunk.setStartTime(bChunk.getStartTime());
            tChunk.setFinishTime(bChunk.getFinishTime());
          }
          // Otherwise, puhs the chunk on the stack if it is a burst
          else if (word_count < BURST_MIN_LENGTH)
          {
            bStack.push(bChunk);
            tStack.push(tChunk);
          }
          // Otherwise, we have just met the burst size, so mark all the words
          // on the stacks with the proper timings.
          else if (word_count == BURST_MIN_LENGTH)
          {
            bStack.push(bChunk);
            tStack.push(tChunk);

            for (Caption.CaptionChunk t, b; !tStack.empty();)
            {
              t = (Caption.CaptionChunk) tStack.pop();
              b = (Caption.CaptionChunk) bStack.pop();
              t.setStartTime(b.getStartTime());
              t.setFinishTime(b.getFinishTime());
            }
          }

          // Iterate to the next word in the caption, incrementing to the
          // next caption if we are at the end of the first.
          if (!bWordItr.hasNext())
          {
            if (burstItr.hasNext())
            {
              burstCaption = (Caption) burstItr.next();
              bWordItr = burstCaption.listIterator(0);
            }
            word_count = 0;
          }

          // Get the next burst chunk
          if (bWordItr.hasNext())
          {
            bChunk = (Caption.CaptionChunk) bWordItr.next();
          }
        }
        // Chunks don't match, so clean things up if there was a match of less
        // than the minimum burst size.
        else
        {
          if (word_count < BURST_MIN_LENGTH && word_count > 0)
          {
            burstItr = capStart;
            bWordItr = wordStart;
            bChunk = (Caption.CaptionChunk) bWordItr.next();
            bStack.clear();
            tStack.clear();
          }

          if (word_count > BURST_MIN_LENGTH)
          {
            tWordItr.previous();
          }

          word_count = 0;
        }
      }
    }

    StatisticsPackage p = transcript.getStatisticsPackage();
    p.add
    (
      new DoubleStatistic
      (
        "Total Processing Time", (finishTime - startTime) / 1000.0 
      )
    );
    p.add(SpeakingRate);
    p.add(UncoveredRate);
    p.add(UncoveredWords);
    p.add(new DoubleStatistic("Speaking Rate", SpeakingRate.getMean()));
    p.add(new IntegerStatistic("Total Words Matched", (int)wordsMatched));
    p.add(new IntegerStatistic("Total Words", (int)totalWords));
    p.add
    (
      new PercentageStatistic
      (
        "Transcript Accuracy", wordsMatched / totalWords
      )
    );

    return transcript;
  }

  /**
   * Creates a timed caption from a burst. Given all the words of a burst and
   * all the words and timings from an entire recognized utterance, extract
   * the timings for those words that are in the burst.
   *
   * @param burst The raw text of a burst of recognized words.
   * @param raw Raw text of owrds in transcript.
   * @param time Text and time-stamp of recognized words as returned by Sphinx.
   *
   * @return A caption with all the words in the burst with their time-stamps.
   *
   */
  private Caption createTimedCaption
  (
    LinkedList burst, 
    LinkedList raw, 
    LinkedList time
  )
  {
    // Create a Caption object to be returned to the caller
    Caption result = new Caption();

    LinkedList timeIndices = 
      new LinkedList(LCS.getShortestEmbedding(burst, raw));

    // Iterator through and find each word that matches
    for 
    (
      ListIterator i = burst.listIterator(), j = timeIndices.listIterator(); 
      i.hasNext() && j.hasNext();
    )
    {
      String burstWord = (String)i.next();
      int timeIdx      = ((Integer)j.next()).intValue();
      
      if(timeIdx >= time.size())
        break;
      
      String timedWord = (String)time.get(timeIdx);
      
      Matcher m = startPat.matcher(timedWord);
      m.find();
      double s = Double.parseDouble(m.group(1));
      
      m = endPat.matcher(timedWord);
      m.find();
      double f = Double.parseDouble(m.group(1));
      
      result.appendCaption(s, f, burstWord);
    }

    int last = ((Integer)timeIndices.getLast()).intValue();

    for(int i = 0; i <= last; ++i)
    {
      raw.removeFirst();
      time.removeFirst();
    }
    
    return result;
  }

  /**
   * Helper function similar to PERL's join function. Returns a string with
   * a space between each object in the collection.
   *
   * @param x The collection of objects to be joined as a string.
   *
   * @return A string with all the objects in x joined as a string.
   *
   */
  private String join(Collection x)
  {
    StringBuffer result = new StringBuffer();

    for (Iterator i = x.iterator(); i.hasNext();)
    {
      result.append(i.next().toString() + " ");
    }

    return result.toString();
  }
}
