//   Copyright (C) 2005 The Regents of the University of California.
//   All Rights Reserved.
//
//   Permission to use, copy, modify, and distribute this software and its
//   documentation for educational, research and non-profit purposes,
//   without fee, and without a written agreement is hereby granted,
//   provided that the above copyright notice, this paragraph and the
//   following three paragraphs appear in all copies.
//
//   Permission to incorporate this software into commercial products may
//   be obtained by contacting the University of California. For
//   information about obtaining such a license contact:
//
//   Chrisanna Waldrop
//   Copyright Officer
//   805-893-7773
//   waldrop@research.ucsb.edu
//
//   This software program and documentation are copyrighted by The Regents
//   of the University of California. The software program and
//   documentation are supplied "as is", without any accompanying services
//   from The Regents. The Regents does not warrant that the operation of
//   the program will be uninterrupted or error-free. The end-user
//   understands that the program was developed for research purposes and
//   is advised not to rely exclusively on the program for any reason.
//
//   IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
//   FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
//   INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
//   ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
//   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
//   CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//   A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
//   BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE
//   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
package edu.ucsb.nmsl.autocap;

import java.util.Vector;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Iterator;
import java.io.FileInputStream;
import java.io.FileOutputStream;

import edu.ucsb.nmsl.tools.*;


/**
 *
 * <p>The responsiblity of this class is to estimate captions that have no time-
 * stamp after the Recognition and Alignment phases of the AutoCap process. The
 * estimation of untime-stamped captions is important because each caption must
 * have a time-stamp in order to create a usable caption file. These estimations
 * are based on the assumption that there are recognized segments of the media
 * file before and after each segment except for the first and last captions.
 * </p>
 *
 * <p>The estmation technique used in the InterCoverageCaptionEstimater is more 
 * accurate than the NaiveCaptionEstimator. This technique uses the local 
 * speaking rate rather than the global speaking rate. The assumption here is 
 * that the speaker does not speak at a consistent rate throughout the video and
 * that accurate estimations of un-time-stamped captions can be made using this 
 * metric. The inter-coverage technique begins at the first word of an un-time-
 * stamped caption and counts the total number of un-time-stamped which includes
 * the first word of the un-time-stamped caption. The time-stamp of the caption 
 * chunks at each end of the unrecognized burst is recorded as well as the total 
 * number of unrecognized words. The difference between the two recorded time-
 * stamps is divided by the number of unrecognized words to calculate the local 
 * speaking rate. This local speaking rate is then multiplied by the number of 
 * words to the nearest recognized word with respect to the first word of the 
 * untime-stamped caption. This result is added to the nearest time-stampe to
 * calculate the estimated time-stampe for the untime-stamped caption. This 
 * process is executed for each untime-stamped caption for a given run of 
 * AutoCap.</p>
 *
 * @author  Allan Knight
 * @version 1.0
 *
 */
public class InterCoverageCaptionEstimator implements CaptionEstimator
{
  /** Global speaking rate of speaker recorded during the recognition phase
      of the AutoCap process. */
  protected double SpeakingRate       = 0;

  /** Estimated beginning time of first caption. Used in case the first caption
      has no time-stamp. */
  protected double EstimatedBeginning = 0;

  /** Estimated finish time of last caption. Used in case the last caption has 
      no time-stamp. */
  protected double EstimatedEnd       = 0;

  /**
   *
   * This constructor creates an instance of the InterCoverageCaptionEstimator
   * with the given global speaking rate. This global speaking rate is only 
   * used for recording statistical information and is not needed for normal
   * opeation of AutoCap.
   *
   * @param rate The global speaking rate of the speaker through out the entire
   *              video.
   *
   */
  public InterCoverageCaptionEstimator(double rate)
  {
    SpeakingRate = rate;
  }
  
  /**
   *
   * This constructor creates an instance of InterCoverageCaptionEstimator with
   * the given spekaing rate and the given estimated start time and end time of
   * the media. This global speaking rate is only used for recording statistical
   * information and is not needed for normal operation of AutoCap.
   * @param rate The global speaking rate of the speaker through out the entire
   *              video.
   * @param estBeg The estimated start time of speaking within the video.
   * @param estEnd The estimated finish time of speaking within the video.   
   *
   */
  public InterCoverageCaptionEstimator
  (
    double rate, 
    double estBeg, 
    double estEnd
  )
  {
    SpeakingRate       = rate;
    EstimatedBeginning = estBeg;
    EstimatedEnd       = estEnd;
  }

  /**
   *
   * <p>This method performs the caption estimation for a given transcript. Each
   * caption of the transcript is investigate, if the first word of the caption
   * has a time-stamp, then no estimation is performed. Otherwise, the time-
   * stamp for the caption is estimated as specified in the class documentation.
   * </p>
   *
   * @param t The Transcript for which the estimation technique will be applied. 
   *
   * @return A Transcript with all captions time-stamped.
   *
   */
  public Transcript completeTranscriptTimes(Transcript t)
  {
    double finishTime = 0.0;

    collectStatistics(t);
    estimateEnds(t);
    
    DataSetStatistic r = 
      (DataSetStatistic)t.getStatisticsPackage().getStatistic
      (
        "Speaking Rate Data"
      );

    DataSetStatistic ur = 
      (DataSetStatistic)t.getStatisticsPackage().getStatistic
      (
        "Uncovered Speaking Rate Data"
      );

    DataSetStatistic uw = 
      (DataSetStatistic)t.getStatisticsPackage().getStatistic
      (
        "Uncovered Words Data"
      );

    double rmin = r.getMean() - 2 * r.getStdDev();

    for (ListIterator curr = t.chunkIterator(0); curr.hasNext();)
    {
      Caption.CaptionChunk c = (Caption.CaptionChunk) curr.next();

      while
      (
        curr.hasNext()           &&
      	c.getStartTime()  >= 0.0 &&
      	c.getFinishTime() >= 0.0
      )
      {
        finishTime = c.getFinishTime();
        c = (Caption.CaptionChunk) curr.next();
      }

      if (!curr.hasNext())
        break;

      int uncoveredCount = 0;
      while
      (
        curr.hasNext()          &&
      	c.getStartTime()  < 0.0 &&
      	c.getFinishTime() < 0.0
      )
      {
        ++uncoveredCount;
        c = (Caption.CaptionChunk) curr.next();
      }
      uw.addDataPoint((double)uncoveredCount);

      int restore = curr.nextIndex();

      double startTime = c.getStartTime();

      double rmax = (startTime - finishTime) / uncoveredCount;
      ur.addDataPoint((double)rmax);

      double delta = 0.0;

      curr.previous();
      c = (Caption.CaptionChunk) curr.previous();
      int j = 1;
      while
      (
        curr.hasPrevious()      &&
      	c.getStartTime()  < 0.0 &&
      	c.getFinishTime() < 0.0
      )
      {
        c.setFinishTime(startTime - delta);
        c.setStartTime(startTime - delta - rmax);
	      c.setMaxError((rmax - rmin) * Math.max(uncoveredCount - j + 1, j));
        delta += rmax;
        c = (Caption.CaptionChunk) curr.previous();
	      ++j;
      }
      curr = t.chunkIterator(restore - 1);      
    }

    return t;
  }

  /**
   *
   * This method estimates the starting time for the first and last captions if
   * no time-stamp has been determined for either. Estimating the ends for the 
   * transcript must be done before all others, as they may depend on the time
   * stamps existing for the estimation to be calculated.
   *
   * @param t The Transcript for which captions are being estimated. 
   *
   */
  private void estimateEnds(Transcript t)
  {
    ListIterator l = t.chunkIterator(0);
    Caption.CaptionChunk c = null;
    int wordCount = 0;

    // Find the first covered chunk in the transcript
    for
    (
      c = (Caption.CaptionChunk) l.next();
      l.hasNext() && c.getStartTime() < 0 && c.getFinishTime() < 0;
      c = (Caption.CaptionChunk) l.next(), ++wordCount
    );
    
    // Fill in the first value
    Caption.CaptionChunk firstChunk = t.getCaption(0).getFirstChunk();
    double s = (c.getStartTime() - EstimatedBeginning) / wordCount;
    firstChunk.setStartTime(EstimatedBeginning);
    firstChunk.setFinishTime(EstimatedBeginning + s);


    // Walk to the end of the transcript
    for (; l.hasNext(); c = (Caption.CaptionChunk) l.next()) ;


    wordCount = 0;
    // Find the last covered chunk in the transcript
    for
    (
      c = (Caption.CaptionChunk) l.previous();
      l.hasPrevious() && c.getStartTime() < 0.0 && c.getFinishTime() < 0.0;
      c = (Caption.CaptionChunk) l.previous(), ++wordCount
    );

    // Fill in the first value
    if(wordCount > 0)
    {
      Caption.CaptionChunk lastChunk = 
	    t.getCaption(t.size() - 1).getLastChunk();
      s = (EstimatedEnd - c.getStartTime()) / wordCount;
      lastChunk.setStartTime(EstimatedEnd - s);
      lastChunk.setFinishTime(EstimatedEnd);
    }
  }

  /**
   *
   * This method is called at the end of the estimation process in order to 
   * collect statistics about how well the speech recognition system performed
   * while transcribing the input media. This method is not necessary for the
   * normal operation of AutoCap.
   *
   * @param t The Transcript of the transcribed input media for which statistics
   *          are being collected. 
   *
   */
  protected void collectStatistics(Transcript t)
  {
    int coverageCount = 0;
    for(Iterator i = t.captionIterator(0); i.hasNext();)
    {
      Caption c = (Caption)i.next();
      for(Iterator j = c.iterator(); j.hasNext();)
      {
        Caption.CaptionChunk ch = (Caption.CaptionChunk)j.next();
        if(ch.getStartTime() >= 0.0 && ch.getFinishTime() >= 0.0)
        {
          ++coverageCount;
          break;
        }
      }
    }

    StatisticsPackage stats = t.getStatisticsPackage();
    stats.add(new IntegerStatistic("Total Segments Covered", coverageCount));
    stats.add(new IntegerStatistic("Total Segments", t.size()));
    stats.add(new PercentageStatistic
    (
      "Segments Coverage", 
      1.0 * coverageCount / t.size())
    );
  }

  /**
   * This method is used for testing purposes only, and is not necessary for the
   * normal operation of AutoCap.
   *
   */
  public static void main(String args[])
  {
    QADTranscriptFileReader r = new QADTranscriptFileReader();

    try
    {
      Transcript t = r.readTranscript(new FileInputStream("captiondata.xml"));

      double time = 0.0;
      for (ListIterator i = t.captionIterator(0); i.hasNext();)
      {
        Caption c = (Caption) i.next();
        Caption.CaptionChunk ch = c.getFirstChunk();
        ch.setStartTime(time);
        ch.setFinishTime(time + 0.33);
        time += 10.0;
      }
      InterCoverageCaptionEstimator est = 
        new InterCoverageCaptionEstimator(0.33);
      est.completeTranscriptTimes(t);
      QADTranscriptFileWriter w = new QADTranscriptFileWriter();
      w.writeTranscript(t, new FileOutputStream("captiondata_copy.xml"));
    } 
    catch (Exception e)
    {
      e.printStackTrace();
      return;
    }
  }
}
