//   Copyright (C) 2005 The Regents of the University of California.
//   All Rights Reserved.
//
//   Permission to use, copy, modify, and distribute this software and its
//   documentation for educational, research and non-profit purposes,
//   without fee, and without a written agreement is hereby granted,
//   provided that the above copyright notice, this paragraph and the
//   following three paragraphs appear in all copies.
//
//   Permission to incorporate this software into commercial products may
//   be obtained by contacting the University of California. For
//   information about obtaining such a license contact:
//
//   Chrisanna Waldrop
//   Copyright Officer
//   805-893-7773
//   waldrop@research.ucsb.edu
//
//   This software program and documentation are copyrighted by The Regents
//   of the University of California. The software program and
//   documentation are supplied "as is", without any accompanying services
//   from The Regents. The Regents does not warrant that the operation of
//   the program will be uninterrupted or error-free. The end-user
//   understands that the program was developed for research purposes and
//   is advised not to rely exclusively on the program for any reason.
//
//   IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
//   FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
//   INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
//   ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
//   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
//   CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//   A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
//   BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE
//   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
package edu.ucsb.nmsl.autocap;

import java.util.Vector;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Iterator;

import edu.ucsb.nmsl.tools.*;

/**
 *
 * <p>The responsiblity of this class is to estimate captions that have no time-
 * stamp after the Recognition and Alignment phases of the AutoCap process. The
 * estimation of untime-stamped captions is important because each caption must
 * have a time-stamp in order to create a usable caption file. These estimations
 * are based on the assumption that there are recognized segments of the media
 * file before and after each segment except for the first and last captions.
 * </p>
 *
 * <p>The estmation technique used in the NaiveCaptionEstimator is very simple. 
 * This technique uses the average speaking rate of the speaker for the entire
 * video. The assumption here is that the speaker speaks at a consistent rate
 * throughout the video and that accurate estimations of un-time-stamped 
 * captions can be made using this metric. The naive technique begins at the 
 * first word of an un-time-stamped caption and counts the number of words to
 * the nearest time-stamped caption chunk. This search simultaneously counts 
 * forward and backward from the starting word until a time-stamped caption
 * chunk. The time-stamp of the nearest recognized word is recorded along with
 * its distance, in words, from the start word. The distance is multiplied by
 * the global speaking rate and added or subtracted from the recorded time-stamp
 * to formulate an estimated time-stamp. This time-stamp is then recorded as the
 * time-stamp for the untime-stamped caption. This process is executed for each
 * untime-stamped caption for a given run of AutoCap.</p>
 *
 * <p>As its name implies, this technique is quite naive and very inaccurate. 
 * The NaiveCaptionEstimator is included only for research purposes because 
 * another technique, the inter-coverate caption estimator technique, performs
 * much better and creates much more accurate time stamps than this technique.
 *
 * @see edu.ucsb.nmsl.autocap.InterCoverageCaptionEstimator
 * 
 * @author  Allan Knight
 * @version 1.0
 *
 */
public class NaiveCaptionEstimator implements edu.ucsb.nmsl.autocap.CaptionEstimator
{
  /** Global speaking rate of speaker recorded during the recognition phase
      of the AutoCap process. */
  protected double SpeakingRate = 0.0;

  /** Vector that holds all the caption that have at least some words in it
      recognized by the speech recognition phase, but not necessarily the
      first word. This vector is used for collecting statistics for research
      purposes only and is not ncessary for normal operator of AutoCap.*/
  protected Vector covered;

  /**
   *
   * This constructor creates an instance of the NaiveCaptionEstimator with the
   * given speaking rate. All estimations are made based on this global speaking
   * rate.
   *
   * @param r The global speaking rate of the speaker throughout the video.
   *
   */
  public NaiveCaptionEstimator(double r)
  {
    SpeakingRate = r;
  }

  /**
   *
   * <p>This method performs the caption estimation for a given transcript. Each
   * caption of the transcript is investigate, if the first word of the caption
   * has a time-stamp, then no estimation is performed. Otherwise, the time-
   * stamp for the caption is estimated as specified in the class documentation.
   * </p>
   *
   * @param t The Transcript for which the estimation technique will be applied. 
   *
   * @return A Transcript with all captions time-stamped.
   *
   */
  public Transcript completeTranscriptTimes(Transcript t)
  {
    // Vector of covered captions
    covered = new Vector();

    // Iterate through each caption and determine if we know their start time
    for(ListIterator capItr = t.captionIterator(0); capItr.hasNext(); )
    {
      // Get the current caption and save the iterator position
      int idx = capItr.nextIndex();
      Caption cap = (Caption)capItr.next();

      // If the start time and finish time of the caption is set 
      if(cap.getTime() >= 0.0 && cap.getFinishTime() >= 0.0)
      {
      	covered.add(new Integer(idx));
      	continue;
      }

      // If the caption does not have a start and finish time, then walk the 
      // list of words in the caption to estimate the start and finish time
      double delta = 0.0;
      ListIterator chunkItr = null;
      for(chunkItr = cap.listIterator(0); chunkItr.hasNext(); )
      {
      	delta += SpeakingRate;
      	Caption.CaptionChunk chunk = (Caption.CaptionChunk)chunkItr.next();
      	if(chunk.getStartTime() >= 0 && chunk.getFinishTime() >= 0)
      	{
      	  // This word has been covered, add it to the list
      	  covered.add(new Integer(idx));
      	  cap.setStartTime(chunk.getStartTime() - delta);
      	  break;
      	}
      }
      
      // Walk to the end of the list to estimate the finish time
      for(; chunkItr.hasNext(); chunkItr.next());

      // Estimate the finish time by walking backwards from the end to the 
      // first word with timing information.
      delta = 0.0;
      for( ; chunkItr.hasPrevious(); )
      {
      	delta += SpeakingRate;
      	Caption.CaptionChunk chunk = (Caption.CaptionChunk)chunkItr.previous();
      	if(chunk.getStartTime() >= 0 && chunk.getFinishTime() >= 0)
      	{
      	  cap.setFinishTime(chunk.getFinishTime() + delta);
      	  break;
      	}
      }
    }

    // We know the number of sentences covered.
    //sentencesCovered = covered.size();
    covered.add(new Integer(Integer.MAX_VALUE));

    // Now go through the list of those captions not covered and estimate
    // the start time based on the closest covered caption.
    int coveredIdx = 0;
    int wordCount  = 0;
    for(ListIterator capItr = t.captionIterator(0); capItr.hasNext(); )
    {
      int currIdx  = capItr.nextIndex();
      
      // Figure out which covered caption the current one is between
      if(currIdx >= ((Integer)covered.elementAt(coveredIdx+1)).intValue())
      {
	      ++coveredIdx;
      }

      Caption cap = (Caption)capItr.next();
      if(cap.getTime() >= 0.0 && cap.getFinishTime() >= 0.0)
      {
      	wordCount = 0;
      	continue;
      }

      // Figure out the closest caption and estimate based on that
      int upperIdx = ((Integer)covered.elementAt(coveredIdx+1)).intValue();
      int lowerIdx = ((Integer)covered.elementAt(coveredIdx)).intValue();

      double time = 0.0;

      // If the closest caption is the later caption, then iterator backwards
      // through each and cound the number of words, otherwise, just use the
      // accumulated words to make the estimation
      if
      (
        capItr.hasNext() && 
	      (currIdx < lowerIdx || currIdx - lowerIdx > upperIdx - currIdx)
      )	      
      {
      	int i = (currIdx < lowerIdx) ? lowerIdx : upperIdx;
      
      	time = t.getCaption(i).getTime();
      	for(; currIdx < i; ++currIdx)
      	  time -= SpeakingRate * t.getCaption(currIdx).size();
      }
      else
      {
	      time = 
	        t.getCaption(lowerIdx).getFinishTime() + wordCount * SpeakingRate;
      }

      // Make the estimation
      cap.setStartTime(time);
      cap.setFinishTime(time + SpeakingRate * cap.size());
      wordCount += cap.size();
    }

    collectStatistics(t);
    return t;
  }

  /**
   * 
   * This method is called at the end of the estimation process in order to 
   * collect statistics about how well the speech recognition system performed
   * while transcribing the input media. This method is not necessary for the
   * normal operation of AutoCap.
   *
   * @param t The Transcript of the transcribed input media for which statistics
   *          are being collected. 
   *
   */
  protected void collectStatistics(Transcript t)
  {
    StatisticsPackage stats = t.getStatisticsPackage();
    stats.add(new IntegerStatistic("Total Segments Covered", covered.size()));
    stats.add(new IntegerStatistic("Total Segments", t.size()));
    stats.add(new PercentageStatistic
      (
        "Segments Coverage", 
        covered.size() / t.size())
      );
  }
}
