//   Copyright (C) 2005 The Regents of the University of California.
//   All Rights Reserved.
//
//   Permission to use, copy, modify, and distribute this software and its
//   documentation for educational, research and non-profit purposes,
//   without fee, and without a written agreement is hereby granted,
//   provided that the above copyright notice, this paragraph and the
//   following three paragraphs appear in all copies.
//
//   Permission to incorporate this software into commercial products may
//   be obtained by contacting the University of California. For
//   information about obtaining such a license contact:
//
//   Chrisanna Waldrop
//   Copyright Officer
//   805-893-7773
//   waldrop@research.ucsb.edu
//
//   This software program and documentation are copyrighted by The Regents
//   of the University of California. The software program and
//   documentation are supplied "as is", without any accompanying services
//   from The Regents. The Regents does not warrant that the operation of
//   the program will be uninterrupted or error-free. The end-user
//   understands that the program was developed for research purposes and
//   is advised not to rely exclusively on the program for any reason.
//
//   IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
//   FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
//   INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
//   ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
//   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
//   CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//   A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
//   BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE
//   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
package edu.ucsb.nmsl.tools;

import javax.swing.BoxLayout;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JTextPane;
import javax.swing.Timer;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.net.URL;
import java.util.Vector;
import javax.media.ControllerEvent;
import javax.media.ControllerListener;
import javax.media.EndOfMediaEvent;
import javax.media.Manager;
import javax.media.Player;
import javax.media.RateChangeEvent;
import javax.media.RealizeCompleteEvent;
import javax.media.StartEvent;
import javax.media.StopEvent;
import javax.media.protocol.DataSource;
import javax.media.protocol.PullBufferDataSource;
import javax.media.protocol.PullDataSource;
import javax.media.protocol.PullSourceStream;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * This class is reponsible for aggregating a Player object and an XML file 
 * containing captions and playing those captions back with synchronized with a 
 * media file. Currently the format of the XML file is that set by QAD, however, 
 * this may change to SMIL in the near future.
 *

 * @version 1.0
 *  
 */
public class CaptionedMediaPlayer extends JPanel
  implements ControllerListener, ActionListener
{
  /** The player object from JMF to play the media file. */
  Player          player       = null;

  /** The JTextPane object that displays the captions. */
  JTextPane       caption      = null;

  /** The Container object that holds the captions and media controls. */
  Container       container    = null;

  /** The DocumentBuilder object used for parsing the caption XML file. */
  DocumentBuilder builder      = null;

  /** The Document object that actually contains the caption XML file. */
  Document        captions     = null;

  /** The list of captions and their times. */
  Vector          captionList  = null;

  /** Index of the current caption. */
  int             captionIndex = 1;

  /** The timer that is used to synchronize the media and the caption text. */ 
  Timer           captionTimer = null;
  
  /*** Static consts to avoid magic numbers and the like ***/

  /** Name of captions element in XML file. Each child of the captions element
      represents an individual caption*/
  protected static final String CAPTIONS_ELEMENT_NAME    = "captions";

  /** Name of the time stamp element in XML file. This element represents the
      time-stamp of a caption */
  protected static final String TIME_ELEMENT_NAME        = "time";

  /** Name of the language element in XML file. This element represents the 
      language of the text of the caption */
  protected static final String LANGUAGE_ELEMENT_NAME    = "language";

  /** Name of the text attribute in XML file. This attribute represents the text 
      of the caption. */
  protected static final String TEXT_ATTRIBUTE_NAME      = "text";

  /** Name of the value attribute in XML file. This attribute represents the 
      text of the caption. */
  protected static final String VALUE_ATTRIBUTE_NAME     = "value";

  /** First element of the caption file. */
  protected static final int    CAPTION_ELEMENT_INDEX    = 0;

  /** Preferred height of the Swing widget that displays the caption. */
  protected static final int    PREFERRED_CAPTION_HEIGHT = 150;

  /** Number of milliseconds per second. */
  protected static final int    MILLISECONDS_PER_SECOND  = 1000;

  /** */
  protected static final double SUPPORTED_MEDIA_RATE     = 1.0;

  /**
   * The default constuctor. Creates all the containers for the GUI. No media 
   * is read yet.
   *

  */
  public CaptionedMediaPlayer()
  {
    // Create the container
    container = new Container();
    container.setLayout(new BoxLayout(container, BoxLayout.Y_AXIS));
    
    // Create the caption text pane
    caption = new JTextPane();
    caption.setEditable(false);
  }
  
  /**
   * Helper method that parses the XML file and stores each caption in a vector.
   * This method creates a new vector each time.
   *
   * @param doc The XML document object (specfically DOM) that contains the
   *            captions.
   *

   *
   */
  protected void parseCaptions(Document doc)
  {
    // Grab all the captions
    NodeList nodes = doc.getElementsByTagName(CAPTIONS_ELEMENT_NAME);
    Node parent = nodes.item(CAPTION_ELEMENT_INDEX);
    
    // Create a new vector to hold the captions
    captionList = new Vector();
    
    // Go through each element until you find a time element
    NodeList times = parent.getChildNodes();
    for(int i = 0; i < times.getLength(); ++i)
    {
      Node item = times.item(i);
      if(item.getNodeName().equals(TIME_ELEMENT_NAME))
      {
      	// Now that we've found a time element, find the lanaguage element. This
      	// contains the text of the caption.
      	NodeList caps = item.getChildNodes();
      	for(int j = 0; j < caps.getLength(); ++j)
      	{
      	  Node n = caps.item(j);
      	  if(n.getNodeName().equals(LANGUAGE_ELEMENT_NAME))
      	  {
      	    // We've found a captions, save it along with the synchronization 
      	    // time.
      	    captionList.add(new Caption
      	    (
      	      item.getAttributes().getNamedItem
      	      (
      	        VALUE_ATTRIBUTE_NAME
      	      ).getNodeValue(),
      	      n.getAttributes().getNamedItem(TEXT_ATTRIBUTE_NAME).getNodeValue()
      	    ));
      	  }
      	}
      }
    }
  }
  
  /**
   * This method is called to play a media file along with its captions. The 
   * media file is any media file accepted by JMF 2.1.1e and the caption file is
   * an XML file with a scheme designed by QAD. This scheme may soon become
   * SMIL.
   *
   * @param p The Player object from JMF that plays the media.
   * @param c Document object that holds the caption information.
   *

   *
   */
  public void play(Player p, Document c)
  {
    try
    {
      // Create a DOM document to parse the caption file
      captions = c;
      
      // Parse the file
      parseCaptions(captions);
      
      // Make sure the player is at least realized
      player = p;
      player.addControllerListener(this);

      int state = player.getState();
      if(state != Player.Prefetched && state != Player.Started)
      {
	      player.start();
      }
    }
    catch(Exception e)
    {
      System.err.println("Got exception " + e);
      e.printStackTrace();
    }
  }

  /**
   * This method is called when the media playing is stopped.
   *

   *
   */
  public void stop()
  {
    assert player != null;

    player.stop();
    captionTimer.stop();
  }

  /**
   * A helper function that finds the caption closest to the time given with in 
   * the media stream. This a down and dirty linear search and allows for the 
   * correct caption to be displayed when the media controls are changed.
   *
   * @param now The current time the media player is at.
   *
   * @return The index of the caption whose time is <= now, -1 otherwise.
   *

   *
   */
  protected int findCaptionIndex(double now)
  {
    // Loop through each one and find the nearest time
    for(int i = 0; i < captionList.size(); ++i)
    {
      if(((Caption)captionList.elementAt(i)).getTime() > now)
	      return i;
    }
    
    // You should never get here.
    return -1;
  }

  /**
   * Listener method that reacts to media events from the player. These events 
   * include realization, start, stop and end of media.
   *
   * @param event The controller event that occurred.
   *

   *
   */
  public synchronized void controllerUpdate(ControllerEvent event)
  {
    // Create the GUI components once the player is realized.
    if(event instanceof RealizeCompleteEvent)
    {
      assert player != null;

      // Get the visual and control components
      Component visComp = player.getVisualComponent();
      Component conComp = player.getControlPanelComponent();
      
      if(visComp != null && conComp != null)
      {
      	// get the width of the mediea, this will be the width of the other
      	// GUI components
      	int width  = (int)visComp.getPreferredSize().getWidth();
      	
      	// Set the preferred size of the captions text pane
      	caption.setPreferredSize
      	(
      	  new Dimension(width, PREFERRED_CAPTION_HEIGHT)
        );
      	
      	// Add the captions and controls to container
      	container.add(caption);
      	container.add(conComp);
      	
      	// Add the visual component and the container to the frame
      	add("Center", visComp);
      	add("South", container);
      	
      	// Calculate the overall height of all three components
      	int height = (int)visComp.getPreferredSize().getHeight() +
      	  (int)caption.getPreferredSize().getHeight() +
      	  (int)conComp.getPreferredSize().getHeight();
      	
      	// Set the height of the frame and validate it.
      	setPreferredSize(new Dimension(width, height));
      	validate();
      }
    }
    // Start the player and start displaying captions. This event occurs when
    // ever the player is started such as at the beginning of the file, after
    // a pause, and when the slider is used to set the position.
    else if(event instanceof StartEvent)
    {
      // Get the current media time
      double offset = player.getMediaTime().getSeconds();

      // Set the captions index
      captionIndex = findCaptionIndex(offset);

      // Create a new timer to synchronize the caption with the media
      captionTimer = new Timer(MILLISECONDS_PER_SECOND * (int)(
        (
	        (Caption)captionList.elementAt(captionIndex)).getTime() - offset),
	        this
	      );
      captionTimer.setRepeats(false);
      captionTimer.start();
    }
    // Stop the timers when the media is stopped. A stop event will occur whev
    // ever the pause button is pushed, or when the slider is selected or when
    // the media is at the end.
    else if(event instanceof StopEvent)
    {						
      captionTimer.stop();
    }
    // This event occurs whenever the rate of the media is changed, either
    // faster or slower. Captions are only displayed during normal play, whic is 
    // rate 1.0
    else if(event instanceof RateChangeEvent)
    {
      if(((RateChangeEvent)event).getRate() != SUPPORTED_MEDIA_RATE)
      {
	      captionTimer.stop();
	      caption.setText("");
      }
    }
    // Clear the captions when the media has reached it end
    else if(event instanceof EndOfMediaEvent)
    {
      caption.setText("");
    }
  }

  /**
   * ActionListener method that is called whenever a timeout occurs. This method 
   * allows for the synchronization of media and captions. However, be careful, 
   * there is no way to differentiate Timer events with other actions. This is a 
   * crude way to synchronize the media and captions, but it will do for now.
   *
   * @param e The ActionEvent corresponding to a timeout (hopefully).
   *

   *
   */
  public void actionPerformed(ActionEvent e)
  {
    // Get the current time
    double offset = player.getMediaTime().getSeconds();

    // Make sure there are more captions left
    if(captionIndex < captionList.size() - 1)
    {
      // Create a new Timer object
      captionTimer = new Timer(MILLISECONDS_PER_SECOND * (int)
        (
	        ((Caption)captionList.elementAt(captionIndex+1)).getTime() - offset),
	        this
	      );
      captionTimer.setRepeats(false);
      captionTimer.start();
    }

    // Set the caption
    caption.setText
      (
        ((Caption)captionList.elementAt(captionIndex++)).getCaption()
      );
  }

  /**
   * Instantiates the CaptionedMediaPlayer and parses the command line arguments 
   * to make sure the right number of arguments are passed in. The first command 
   * line argument is assumed to be the media file, and the second the caption
   * file in the form of an XML file.
   *
   * @param argv The command line arguments
   *

   *
   */
  public static void main(String [] argv)
  {
    if(argv.length < 2)
    {
      System.err.println("Usage: java MediaTestApplet <media file> <captions>");
      System.exit(-1);
    }

    final CaptionedMediaPlayer player = new CaptionedMediaPlayer();
    JFrame frame = new JFrame();
    frame.getContentPane().add(player);
    frame.addWindowListener
      (
        new WindowAdapter()
	      {
	        public void windowClosing(WindowEvent e)
	        {
	          player.stop();
	          System.exit(0);
	        }
	      }
      );
    frame.show();

    try
    {
      DataSource s = Manager.createDataSource(new File(argv[0]).toURL());
      Player     p = Manager.createPlayer(s); 

      Document d = 
	      DocumentBuilderFactory.newInstance().newDocumentBuilder().parse
	      (
	        new File(argv[1])
	      );

      player.play(p, d);
      frame.setSize(player.getPreferredSize());
    }
    catch(Exception e)
    {
      System.err.println("Got exception " + e);
      e.printStackTrace();
      System.exit(-1);
    }
  }
}
