//   Copyright (C) 2005 The Regents of the University of California.
//   All Rights Reserved.
//
//   Permission to use, copy, modify, and distribute this software and its
//   documentation for educational, research and non-profit purposes,
//   without fee, and without a written agreement is hereby granted,
//   provided that the above copyright notice, this paragraph and the
//   following three paragraphs appear in all copies.
//
//   Permission to incorporate this software into commercial products may
//   be obtained by contacting the University of California. For
//   information about obtaining such a license contact:
//
//   Chrisanna Waldrop
//   Copyright Officer
//   805-893-7773
//   waldrop@research.ucsb.edu
//
//   This software program and documentation are copyrighted by The Regents
//   of the University of California. The software program and
//   documentation are supplied "as is", without any accompanying services
//   from The Regents. The Regents does not warrant that the operation of
//   the program will be uninterrupted or error-free. The end-user
//   understands that the program was developed for research purposes and
//   is advised not to rely exclusively on the program for any reason.
//
//   IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
//   FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
//   INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
//   ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
//   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
//   CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//   A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
//   BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE
//   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
package edu.ucsb.nmsl.tools;

import java.util.LinkedList;
import java.util.Collection;
import java.util.ListIterator;
import java.util.Arrays;

/**
 *
 * <p>This class represents a statistic that represents a set of statistical
 * information. Using a set of statistics allows for the calculation of an 
 * average, median, mode and standard deviation, all of which are not possible
 * using the StatisticalPackage class.</p>
 *
 * <p>The purpose of this class is to facilitate the research associated with
 * the creation of the system and is not needed in the normal operation of
 * AutoCap. Only the displayString method is overloaded withing the definition.
 * </p>
 *

 * @version 1.0
 *
 */
public class DataSetStatistic extends BaseStatistic
{
  /** A linked list containing all the data in the data set. */
  protected LinkedList Data;
  
  /** A running sum of all the data in the data set. */
  protected double     Sum = 0;
  
  /** A running variance of all the data in the data set. */
  protected double     Var = 0;

  /**
   *
   * This constructor creates a DataSetStatistic with a given name and with data
   * from the specified collection.
   *
   * @param name The name associated with the instance.
   * @param c A collection of data that becomes the data points in the data set.
   *
   */
  public DataSetStatistic(String name, Collection c)
  {
    Name = new String(name);
    Data = new LinkedList();
    Data.addAll(c);
  }

  /**
   *
   * This contructor builds a DataSetStatistic instance with the name 
   * specified by "name" with an empty data set.
   *
   * @param name The name of the DataSetStatisic.
   *

   *
   */
  public DataSetStatistic(String name)
  {
	  Name = new String(name);
	  Data = new LinkedList();
  }
    
  /**
   *
   * This method adds a data point to the set. All data is entered as a double.
   *
   * @param x The data point to be added to the data set. 
   *

   *
   */
  public void addDataPoint(double x)
  {
    Data.add(new Double(x));
    Sum += x;
    Var += x * x;
  }
    
  /**
   *
   * This method removes a data point from the set.
   *
   * @param x The value of the data point to be removed. 
   *

   *
   */
  public void removeDataPoint(double x)
  {
    Data.remove(new Double(x));
    Sum -= x;
    Var -= x * x;
  }

  /**
   *
   * This method calculates are returns the standard deviation of all the data
   * in the set. The calculation is sped up by keeping a running sum and 
   * variance as data points are added and subtracted from the set.
   *
   * @return The standard deviation of all the data points in the set.
   *

   *
   */
  public double getStdDev()
  {
    int size = Data.size();
    return Math.sqrt((Var - Sum * Sum / size) / size);
  }

  /**
   * 
   * This method returns the mean of all data points in the set. The calculation
   * is sped up by using a running sum of all the data points as they are added
   * to and subracted from the set.
   *
   * @return The mean of all data points in the set.
   *

   *
   */
  public double getMean()
  {
    return Sum / Data.size();
  }
    
  /**
   *
   * This method calculates and returns the median value of the data points in 
   * the set. This is the straight forward linear implementation of the median
   * algorithm. Perhaps in the future a faster median algorithm can be used.
   *
   * @return The median value of the data points in the set.
   *

   *
   */
  public double getMedian()
  {
    Object data[] = Data.toArray();
    Arrays.sort(data);
    double median = 0.0;

    if(data.length % 2 == 0)
    {
      median = 
        (
          ((Double)data[data.length / 2 - 1]).doubleValue() + 
          ((Double)data[data.length / 2]).doubleValue()
        ) / 2;
    }
    else
    {
      median = ((Double)data[data.length / 2 - 1]).doubleValue();
    }

    return median;
  }

  /**
   *
   * This method returns the mode of all the data points in the set. There is 
   * no implementation yet as this functionality was not needed during 
   * the implemenation of AutoCap.
   *
   * @return 0.0
   *

   *
   */
  public double getMode()
  {
    return 0.0;
  }

  /**
   * This method returns a string that can be used to neatly show the 
   * statistical information associated with this particular instance. Each data
   * point in the set is displayed on an individual line and the central 
   * tendencies are listed at the end.
   *
   * @return A string representing the statistical information. This string 
   *         usually takes the form of &lt;Name&gt; = &lt;Data Point1&gt; ...;     
   *

   *
   */
  public String displayString()
  {
    StringBuffer display = new StringBuffer(Name + "(" + Data.size() + ") =\n");
    for(ListIterator i = Data.listIterator(); i.hasNext();)
    {
        Double d = (Double)i.next();
        display.append(d + "\n");
    }
    display.append("StdDev = " + getStdDev() + "\n");
    display.append("Mean   = " + getMean()   + "\n");
    display.append("Median = " + getMedian() + "\n");
    display.append("Mode   = " + getMode()   + "\n");
    
    return display.toString();
  }

  /**
   * 
   * This method is meant for testing purposes and is not call in the operation
   * of AutoCap.
   *

   *
   */
  public static void main(String args[])
  {
    DataSetStatistic captionError = new DataSetStatistic("CaptionError");
    captionError.addDataPoint(-3.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(-1.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(-1.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(-1.0);
    captionError.addDataPoint(-1.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(-1.0);
    captionError.addDataPoint(-2.0);
    captionError.addDataPoint(-1.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(0.0);
    captionError.addDataPoint(1.0);
    
    System.out.println(captionError.displayString());
  }
}

