//   Copyright (C) 2005 The Regents of the University of California.
//   All Rights Reserved.
//
//   Permission to use, copy, modify, and distribute this software and its
//   documentation for educational, research and non-profit purposes,
//   without fee, and without a written agreement is hereby granted,
//   provided that the above copyright notice, this paragraph and the
//   following three paragraphs appear in all copies.
//
//   Permission to incorporate this software into commercial products may
//   be obtained by contacting the University of California. For
//   information about obtaining such a license contact:
//
//   Chrisanna Waldrop
//   Copyright Officer
//   805-893-7773
//   waldrop@research.ucsb.edu
//
//   This software program and documentation are copyrighted by The Regents
//   of the University of California. The software program and
//   documentation are supplied "as is", without any accompanying services
//   from The Regents. The Regents does not warrant that the operation of
//   the program will be uninterrupted or error-free. The end-user
//   understands that the program was developed for research purposes and
//   is advised not to rely exclusively on the program for any reason.
//
//   IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
//   FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
//   INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
//   ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
//   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
//   CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//   A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
//   BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE
//   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
package edu.ucsb.nmsl.tools;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;
import java.util.TreeMap;
import java.util.List;
import java.util.Arrays;
import java.util.LinkedList;

/**
 *
 * <p>This class is responsible for finding the <i>Longest Common Subsequence
 * </i> (LCS) of a set. A subsequence of a set is defined as sequence, <b>B</b> 
 * of obtained by removing elements of sequence <b>A</b>. For example, if <b>A
 * </b> = { 'A', 'B', 'C', 'B', 'D', 'A', 'B' } then {'A', 'B', 'B', 'B'} is a 
 * subsequence of <b>A</b> if you remove items 3, 5, 6 from the sequence. 
 * Furthermore, the longest common subsequence is a subsequence of two
 * sequences containing the greatest number of common elements between the two 
 * sequences. Two sequences may have multiple common subsequences that are of 
 * the lognest length. The implementation provide in this class is based on the
 * classic <I>Dynamic Programming</I> approach and achieves O(n<sup>2</sup>)
 * performance.</p>
 *
 * <p>The LCS class is capable of calculating an LCS for strings as well as
 * collections of objects. Similarly, the sets returned by the LCS class are
 * either collections of strings or collections of objects. Also, this 
 * implementation of the LCS offers a method called "getShortestEmbedding()".
 * An embedding is defined a sequence of indices for a longest common 
 * subsequence. For example, an embedding, with respect to the first sequence,
 * for the sequences { 'B', 'A', 'B' }, and { 'A', 'B', 'A' } is { 1, 2 } since 
 * { 'B', 'A' } is an LCS for the two sequences. Another embedding is { 2, 3 }
 * since { 'A', 'B' } is also an LCS. The shortest embedding, however, is 
 * defined as the embedding with the shortest difference between the first and
 * last index of an embedding. Again, it is possible for two sequences to have
 * more than one shortest embedding, just as they can have more than one LCS.
 * </p>
 * 
 * <p>The LCS algorithm is important to the alignment phase of the AutoCap
 * process. Using this algorithm AutoCap is able to filter out words that do 
 * not match between what is recognized and what is actually in the transcript.
 * Once all utterances have been collected during the speech recognition phase,
 * the LCS is calculated between all the utterances and the transcript. The 
 * result is all the common words between the two, with the order retained. The
 * time-stamps for the recognized words are then used to determine the time-
 * stamp of each caption withing the transcript.</p>
 *
 * <p>Future implementations of this algorithm could be done using the either 
 * the Hirschberg or Hunt-Szymanski Paradigm. In the case of AutoCap, the Hunt-
 * Szymanski paradigm may be preferred since the Hirschberg paradigm can degrade
 * to worse than O(n<sup>2</sup>) performance if the number of matches is not
 * small compared to the size of each subsequence. Since this is not likely when
 * using AutoCap, the Hunt-Szymanski paradigmn will perform better for this 
 * application</p>
 *
 * @author Allan Knight
 * @version 1.0
 *
 */
public class LCS
{
  /** Distance matrix used to hold the lengths of the LCS */
  protected static int c[][];

  /** Print matrix used to reconstruct an LCS of two sequences */
  protected static int b[][];

  /** Indicates a move should be diagnal within the Print matrix */
  protected static final int DIAG    = 0;

  /** Indicates a move should be up within the Print matrix */
  protected static final int UP      = 1;

  /** Indicates a move should be to the left within the Print matrix */
  protected static final int LEFT    = 2;

  /** Symbols used to print the Print matrix for debuggin purposes */
  protected static final char SYMBOLS[] = {'\\', '|', '_', '>'};

  /**
   *
   * This method computes the LCS of two collections and returns it as a 
   * collection.
   *
   * @param x A sequence that will be compared to y to find their LCS.
   * @param y A sequence that will be compared to x to find their LCS. 
   *
   * @return A collection of objects that represent the LCS of x and y.
   *
   */
  public static Collection getLCS(Collection x, Collection y)
  {
    populateMatrix(x, y);
    
    return buildLCS(x);
  }

  /**
   *
   * This method computes the LCS of two stringss and returns it as a string. 
   * The LCS is calculated with respect to the words within the string, not the
   * characters.
   *
   * @param x A String that will be compared to y to find their LCS.
   * @param y A String that will be compared to x to find their LCS. 
   *
   * @return A String that represent the LCS of x and y.
   *
   */
  public static String getWordLCS(String x, String y)
  {
    Collection lcs = getLCS
      (
        Arrays.asList(x.split("\\s+")), 
	      Arrays.asList(y.split("\\s+"))
      );

    StringBuffer strLCS = new StringBuffer();
    for(Iterator i = lcs.iterator(); i.hasNext();)
    {
      strLCS.append((String)i.next());
      if(i.hasNext())
      {
	      strLCS.append(" ");
      }
    }

    return strLCS.toString();
  }

  /**
   * 
   * This method returns an embedding with respect to x for the sequences x and
   * y.
   * 
   * @param x A sequence that will be compared to y to find their LCS.
   * @param y A sequence that will be compared to x to find their LCS. 
   *
   * @return A collection of Integers that contain the indices of a LCS with
   *         respect to x between subsequences x and y.
   *
   */
  public static Collection getLCSIndices(Collection x, Collection y)
  {
    Vector lcs = new Vector();
    populateMatrix(x, y);

    getLCSIndices(lcs, c.length-1, c[0].length-1);

    return lcs;
  }

  /**
   *
   * This method returns the shortest embedding with respect to x between 
   * subsequences x and y starting from startI and startJ the distance matrix.
   * getShortestEmbeddingFrom is a helper method for getShortestEmbedding and
   * should not be called anywhere else in your code.
   *
   * @param x A sequence that will be compared to y to find the shortest 
   *          embedding.
   * @param y A sequence that will be compared to x to find the shortest
   *          embedding.
   * @param startI The starting row in the distance matrix for this call. 
   * @param startJ The starting column in the distance matrix for this call. 
   *
   * @return A collection of indices for the shortest embedding for the 
   *         indicated subsequences of x and y.
   *
   */
  private static Collection getShortestEmbeddingFrom
  (
    Collection x, 
    Collection y, 
    int startI, 
    int startJ
  )
  {
    LinkedList rest = new LinkedList();
    int curr = c[startI][startJ];
    int p    = c[x.size()][y.size()];

    ++startJ; ++startI;

    for(int j = startJ; j <= y.size() && startI <= x.size(); ++j)
    {
      int tail    = c[x.size() - 0][j];

      if(tail == curr)
        continue;

      int i;
      for
      (
        i = startI; 
        i < x.size() && b[i][j] != DIAG && x.size() - i >= p - curr; 
        ++i
      );
      
      if(b[i][j] == DIAG)
      {
      	rest.add(new Integer(j - 1));
      	curr++;
      	startI = i + 1;
      }
    }

    return rest;
  }

  /**
   *
   * This method returns the shortest embedding with respect to x between 
   * subsequences x and y.
   *
   * @param x A sequence that will be compared to y to find the shortest 
   *          embedding.
   * @param y A sequence that will be compared to x to find the shortest
   *          embedding.
   *
   * @return A collection of indices for the shortest embedding for the 
   *         subsequences of x and y.
   *
   */
  public static Collection getShortestEmbedding(Collection x, Collection y)
  {
    LinkedList embedding = new LinkedList();
    LinkedList curr      = new LinkedList();

    populateMatrix(x, y);

    int p = c[x.size()][y.size()];

    for(int i = 1; i <= x.size(); ++i)
    {
      for(int j = 1; j <= y.size() && c[i][j] <= 1; ++j)
      {
      	if(b[i][j] == DIAG)
      	{
      	  curr.clear();
      	  curr.add(new Integer(j - 1));
      	  curr.addAll(getShortestEmbeddingFrom(x, y, i, j));
      
      	  if(embedding.isEmpty())
      	  {
      	    embedding.addAll(curr);
      	    continue;
      	  }
      	  
      	  int currSize = 
      	    ((Integer)curr.getLast()).intValue() - 
      	    ((Integer)curr.getFirst()).intValue();
      	  
      	  int embSize =
      	    ((Integer)embedding.getLast()).intValue() -
      	    ((Integer)embedding.getFirst()).intValue();
      	  
      	  if(currSize < embSize && curr.size() == p)
      	  {
      	    embedding.clear();
      	    embedding.addAll(curr);
      	  }  
      	}
      }
    }

    return (embedding.size() == p) ? embedding : new LinkedList();
  }

  /**
   *
   * This method is a helper function that computes the indices of an LCS for a
   * given collection.
   *
   * @param col A collection that holds the indices of an embedding. 
   *
   */
  protected static void getLCSIndices(Collection col, int i, int j)
  {
    if(i <= 0 || j <= 0)
      return;

    if(b[i][j] == DIAG)
    {
      getLCSIndices(col, i - 1, j - 1);
      col.add(new Integer(j-1));
    }
    else if(b[i][j] == LEFT)
      getLCSIndices(col, i, j - 1);
    else
      getLCSIndices(col, i - 1, j);
  }

  /**
   *
   * This method is a helper function that recursively builds the actual LCS for
   * two given collections.
   *
   * @param x One of the collections for which the LCS is being calculated.
   *
   * @return The collection that holds the LCS after it is built.
   *
   */
  protected static Collection buildLCS(Collection x)
  {
    Vector lcs = new Vector();
    Object xi[] = x.toArray();

    buildLCS(xi, lcs, c.length-1, c[0].length-1);

    return lcs;
  }

  /**
   *
   * This recursive method is called by buildLCS and does the actual work of 
   * building the LCS from the print matrix.
   *
   * @param x An array of Objects that is one of the collections for which the
   *          LCS is being calculated.
   * @param col The collection that holds the LCS after it is built.
   * @param i The index of the row in the print matrix that is currently being
   *          inspected.
   * @param j The index of the column in the print matrix that is currently
   *          being used inspected.
   *
   */
  protected static void buildLCS(Object x[], Collection col, int i, int j)
  {
    if( i <= 0 || j <= 0)
      return;

    if(b[i][j] == DIAG)
    {
      buildLCS(x, col, i - 1, j - 1);
      col.add(x[i-1]);
    }
    else if(b[i][j] == LEFT)
      buildLCS(x, col, i, j - 1);
    else
      buildLCS(x, col, i - 1, j);
  }

  /**
   *
   * This method is called to popluate both the distance and print matrix. These
   * matrices are used in the construction of an LCS between two given 
   * subsequences.
   *
   * @param x A sequence that will be compared to y to find their LCS.
   * @param y A sequence that will be compared to x to find their LCS. 
   *
   */
  protected static void populateMatrix(Collection x, Collection y)
  {
    c = new int[x.size()+1][y.size()+1];
    b = new int[x.size()+1][y.size()+1];

    Object xi [] = x.toArray();
    Object yj [] = y.toArray();

    for(int i = 0; i <= x.size(); ++i)
    {
      c[i][0] = 0;
    }

    for(int i = 0; i <= y.size(); ++i)
    {
      c[0][i] = 0;
    }

    for(int i = 1; i <= x.size(); ++i)
    {
      for(int j = 1; j <= y.size(); ++j)
      {
      	if(xi[i-1].equals(yj[j-1]))
      	{
      	  c[i][j] = c[i-1][j-1] + 1;
      	  b[i][j] = DIAG;
      	}
      	else 
      	{
      	  c[i][j] = Math.max(c[i][j-1], c[i-1][j]);
      	  if(c[i-1][j] > c[i][j-1])
      	    b[i][j] = UP;
      	  else
      	    b[i][j] = LEFT;
      	}
      }
    }
  }

  /**
   * This main method is for testing and debugging purposes. This method is not
   * called during the normal operation of LCS nor AutoCap.
   *
   */
  public static void main(String args[])
  {
    String n = 
      "the meant that the set or want in the based on some minimum container" +
      " size";
  String e = "on some minimum container size";

    List x = Arrays.asList(e.split("\\s+"));
    List y = Arrays.asList(n.split("\\s+"));

    Collection c = LCS.getShortestEmbedding(x, y);

    for(Iterator i = c.iterator(); i.hasNext(); )
    {
      Integer idx = (Integer)i.next();
      System.out.println(idx.intValue() + ", ");
    }
    System.out.println();
  }
}
