//   Copyright (C) 2005 The Regents of the University of California.
//   All Rights Reserved.
//
//   Permission to use, copy, modify, and distribute this software and its
//   documentation for educational, research and non-profit purposes,
//   without fee, and without a written agreement is hereby granted,
//   provided that the above copyright notice, this paragraph and the
//   following three paragraphs appear in all copies.
//
//   Permission to incorporate this software into commercial products may
//   be obtained by contacting the University of California. For
//   information about obtaining such a license contact:
//
//   Chrisanna Waldrop
//   Copyright Officer
//   805-893-7773
//   waldrop@research.ucsb.edu
//
//   This software program and documentation are copyrighted by The Regents
//   of the University of California. The software program and
//   documentation are supplied "as is", without any accompanying services
//   from The Regents. The Regents does not warrant that the operation of
//   the program will be uninterrupted or error-free. The end-user
//   understands that the program was developed for research purposes and
//   is advised not to rely exclusively on the program for any reason.
//
//   IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
//   FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
//   INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
//   ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
//   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
//   CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//   A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
//   BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE
//   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
package edu.ucsb.nmsl.tools;

import java.io.InputStream;
import java.io.FileInputStream;
import java.util.LinkedList;
import java.util.ListIterator;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 * <p>This class is an implementation of the TranscriptFileReader interface that
 * is used to read transcripts from disk in the format used for educational 
 * materials created by QAD, Inc. This format is specific to their needs and
 * is in no way derived from any other standard used for storing multimedia
 * meta data. </p>
 * 
 * <p>The QAD caption file format consists of an XML file that is capable of 
 * storing the captions along with the time the caption is spoken in an 
 * associated video. These captions can be in multiple languages. The folling
 * is an example of what is contained withing a QAD caption file:</p>
 *
 * <pre>
 &lt;xml&gt;
   &lt;captions&gt;
     &lt;time value="00:00:02"&gt;
       &lt;language="English" text="caption 1"/&gt;
     &lt;/time&gt; 
     &lt;time value="00:00:10"&gt;
       &lt;language="English" text="caption 2"/&gt;
     &lt;/time&gt; 
   &lt;/captions&gt;
 &lt;/xml&gt;
 * </pre>
 * 
 * <p>The above XML represents the captions as a node called "caption" with 
 * multiple child nodes called "time". Each time node represents a caption that
 * has multiple "language" children. These "language" children represent the
 * caption text that is to be displayed for a particular language. One 
 * "langauge" element is diplayed based on the preferred language of the 
 * viewer of the material.</p>
 *
 *<p>Other formats can be used by AutoCap by other implementations of the 
 * TranscriptFileReader interface.</p>
 *  

 * @version 1.0
 *
 */

public class QADTranscriptFileReader implements TranscriptFileReader
{
  /**
   *
   * This default constructor creates a default instance of the 
   * QADTranscriptFileReader class.
   *

   *
   */
   public QADTranscriptFileReader()
  {
  }

  /**
   *
   * This method accepts an InputStream instance to read captions and time 
   * stamps from in order to create an instance of a Transcript object.
   *
   * @param in - An InputStream instance that Transcript information will be 
   *             read from. 
   *            
   * @return A newly created Transcript instance created based on the 
   *         information read from the InputStream.
   *

   *
   */
  public Transcript readTranscript(InputStream in)
  {
    // Create the document
    Document d;
    try
    {
      d = 
	DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
    }
    catch(Exception e)
    {
      e.printStackTrace();
      return null;
    }

    return readTranscript(d);
  }

  /**
   *
   * This method accepts a Document instance to read captions and time stamps
   * from in order to create an instance of a Transcript object. The Document
   * object is defined as part of the default XML DOM implementation that 
   * accompanies the JVM from Sun.
   *
   * @param d - An instance of a Document object that contains the caption
   *            information in an XML DOM object.
   *           
   * @return A newly created Transcript instance created based on the 
   *         information read from the InputStream.
   *

   *
   */
  public Transcript readTranscript(Document d)
  {
    Element root = d.getDocumentElement();
    NodeList s = root.getElementsByTagName("language");
    
    LinkedList transcript = new LinkedList();

    for(int i = 0; i < s.getLength(); ++i)
    {
      // Get each item in the XML file
      Node n = s.item(i);

      // Break the caption into all it words, remove the punctuation, and
      // convert all characters to lowercase and remove whitspace.
      String[] words = 
    	n.getAttributes().
    	getNamedItem("text").
    	getNodeValue().
    	replaceAll("[#$%&()*+\\-,./:;<=>?@[\\\\]]", " ").
    	replaceAll("\\s+", " ").
    	split("\\s+");

      // Create a caption
      Caption newCaption = null;
      if(words.length > 0)
      {
      	newCaption = new Caption();
      	transcript.add(newCaption);
      }

      // Add each word to the caption, then we can iterate through each word
      for(int j = 0; j < words.length; ++j)
      {
      	if(words[j].length() > 0)
      	  newCaption.appendCaption(-1.0, -1.0, words[j]);
      }
    }

    return new Transcript(transcript);
  }

  /**
   *
   * This method is for testing purposes only. This method is not used in the 
   * normal operation of AutoCap. 
  *

   *
   */
  public static void main(String args[])
  {
    try
    {
      QADTranscriptFileReader reader = new QADTranscriptFileReader();
      Transcript t = reader.readTranscript(new FileInputStream(args[0]));
      for(ListIterator i = t.captionIterator(0); i.hasNext(); )
      {
      	Caption c = (Caption)i.next();
      	System.out.println(c.getCaption());
      }

      System.out.println(t.getAsText());
    }
    catch(Exception e)
    {
      e.printStackTrace();
      return;
    }
  }
}
