//   Copyright (C) 2005 The Regents of the University of California.
//   All Rights Reserved.
//
//   Permission to use, copy, modify, and distribute this software and its
//   documentation for educational, research and non-profit purposes,
//   without fee, and without a written agreement is hereby granted,
//   provided that the above copyright notice, this paragraph and the
//   following three paragraphs appear in all copies.
//
//   Permission to incorporate this software into commercial products may
//   be obtained by contacting the University of California. For
//   information about obtaining such a license contact:
//
//   Chrisanna Waldrop
//   Copyright Officer
//   805-893-7773
//   waldrop@research.ucsb.edu
//
//   This software program and documentation are copyrighted by The Regents
//   of the University of California. The software program and
//   documentation are supplied "as is", without any accompanying services
//   from The Regents. The Regents does not warrant that the operation of
//   the program will be uninterrupted or error-free. The end-user
//   understands that the program was developed for research purposes and
//   is advised not to rely exclusively on the program for any reason.
//
//   IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
//   FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
//   INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
//   ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
//   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
//   CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//   A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
//   BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE
//   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
package edu.ucsb.nmsl.tools;

import java.io.OutputStream;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.util.Iterator;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.dom.DOMSource;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 * <p>This class is an implementation of the TranscriptFileWriter interface that
 * is used to write transcripts to disk in the format used for educational 
 * materials created by QAD, Inc. This format is specific to their needs and
 * is in no way derived from any other standard used for storing multimedia
 * meta data. </p>
 * 
 * <p>The QAD caption file format consists of an XML file that is capable of 
 * storing the captions along with the time the caption is spoken in an 
 * associated video. These captions can be in multiple languages. The folling
 * is an example of what is contained withing a QAD caption file:</p>
 *
 * <pre>
 &lt;xml&gt;
   &lt;captions&gt;
     &lt;time value="00:00:02"&gt;
       &lt;language="English" text="caption 1"/&gt;
     &lt;/time&gt; 
     &lt;time value="00:00:10"&gt;
       &lt;language="English" text="caption 2"/&gt;
     &lt;/time&gt; 
   &lt;/captions&gt;
 &lt;/xml&gt;
 * </pre>
 * 
 * <p>The above XML represents the captions as a node called "caption" with 
 * multiple child nodes called "time". Each time node represents a caption that
 * has multiple "language" children. These "language" children represent the
 * caption text that is to be displayed for a particular language. One 
 * "langauge" element is diplayed based on the preferred language of the 
 * viewer of the material.</p>
 *
 *<p>Other formats can be used by AutoCap by other implementations of the 
 * TranscriptFileWriter interface.</p>
 *  

 * @version 1.0
 *
 */
public class QADTranscriptFileWriter implements TranscriptFileWriter
{
  /**
   *
   * This default constructor creates a default instance of the 
   * QADTranscriptFileWriter class.
   *

   *
   */
  public QADTranscriptFileWriter()
  {
  }

  /**
   *
   * This method accepts a Transcript object in order for the caption and time-
   * stamps to be written to an OutputStream.
   *
   * @param t - A Transcript instance to be written.
   * @param out - An OutputStream instance to write the Transcript instance to.
   *

   *
   */
  public void writeTranscript(Transcript t, OutputStream out)
  {
    // Build document factory and eventually the document itself
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    try
    {
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document        doc     = builder.newDocument();

      // Create the captions field
      Element captions = (Element)doc.createElement("captions");
      doc.appendChild(captions);

      // Create all the time fields of the caption file
      for(Iterator i = t.captionIterator(0); i.hasNext(); )
      {
	      Caption cap = (Caption)i.next();

	      Element time = (Element)doc.createElement("time");
	      captions.appendChild(time);

	      // Set the value attribute, that has the timing we extracted.
	      time.setAttribute
	      (
	        "value", 
	        createTimeString((int)Math.round(cap.getTime()) - 1)
	      );
	
     	  // Add the language field that contains the text of the caption.
     	  Element language = (Element)doc.createElement("language");
     	  time.appendChild(language);
        
     	  // Set the caption text
     	  language.setAttribute("name", "English");
     	  language.setAttribute("text", cap.getCaption());
      }

      // Transform the DOM object into a file stream and save it
      TransformerFactory tFactory    = TransformerFactory.newInstance();
      Transformer        transformer = tFactory.newTransformer();

      DOMSource source = new DOMSource(doc);
      StreamResult result = new StreamResult(out);
      transformer.transform(source, result);
    }
    catch(Exception e)
    {
      e.printStackTrace();
    }
  }

  /**
   *
   * Helper function that converts and integer, representing the number of 
   * seconds, into a string of the format HH:MM:SS. Used to print to the XML
   * file.
   *
   * @param value - The integer representation of time in seconds.
   *
   * @return A string representation of the time.
   *

   *
   */
  protected String createTimeString(int value)
  {
    int sec = value % 60;
    int min = (value / 60) % 60;
    int hr  = value / 3600;

    return new String
      (
        ((hr  < 10) ? "0" + hr  : "" + hr ) + ":" + 
	((min < 10) ? "0" + min : "" + min) + ":" +
	((sec < 10) ? "0" + sec : "" + sec)
      );
  }

  /**
   *
   * This method is for testing purposes only. This method is not used in the 
   * normal operation of AutoCap. 
   *

   *
   */
  public static void main(String args[])
  {
    try
    {
      QADTranscriptFileReader r = new QADTranscriptFileReader();
      Transcript t = r.readTranscript(new FileInputStream("captiondata.xml"));
      
      QADTranscriptFileWriter w = new QADTranscriptFileWriter();
      w.writeTranscript(t, new FileOutputStream("captiondata_copy.xml"));
    }
    catch(Exception e)
    {
      e.printStackTrace();
      return;
    }
  }
}
